#include "Estimator.h"
#include "Operator.h"
#include "DOFMatrix.h"
#include "DOFVector.h"
#include "Assembler.h"
#include "Traverse.h"
#include "Parameters.h"

namespace AMDiS {

  void r(const ElInfo              *elInfo,
	 int                        numPoints,
	 const double              *uhIq,
	 const WorldVector<double> *grdUhIq,
	 const WorldMatrix<double> *D2UhIq,
	 const double              *uhOldIq,
	 const WorldVector<double> *grdUhOldIq,
	 const WorldMatrix<double> *D2UhOldIq,
	 DOFMatrix *A, 
	 DOFVector<double> *fh,
	 Quadrature *quad,
	 double *result)
  {
    ::std::vector<Operator*>::iterator it;
    ::std::vector<double*>::iterator fac;
    double factor;

    // lhs
    for(it = const_cast<DOFMatrix*>(A)->getOperatorsBegin(),
	  fac = const_cast<DOFMatrix*>(A)->getOperatorEstFactorBegin(); 
	it != const_cast<DOFMatrix*>(A)->getOperatorsEnd(); 
	++it, ++fac) 
      {
	factor = *fac ? **fac : 1.0;
	if (factor) {
	  if (D2UhIq) {
	    (*it)->evalSecondOrder(numPoints, uhIq, grdUhIq, D2UhIq, result, -factor);
	  }

	  if (grdUhIq) {
	    (*it)->evalFirstOrderGrdPsi(numPoints, uhIq, grdUhIq, D2UhIq, result, factor);
	    (*it)->evalFirstOrderGrdPhi(numPoints, uhIq, grdUhIq, D2UhIq, result, factor);
	  }

	  if (uhIq) {
	    (*it)->evalZeroOrder(numPoints, uhIq, grdUhIq, D2UhIq, result, factor);
	  }
	}
      }

    // rhs
    for(it = const_cast<DOFVector<double>*>(fh)->getOperatorsBegin(),
	  fac = const_cast<DOFVector<double>*>(fh)->getOperatorEstFactorBegin(); 
	it != const_cast<DOFVector<double>*>(fh)->getOperatorsEnd(); 
	++it, ++fac) 
      {
	factor = *fac ? **fac : 1.0;
	if(factor) {
	  if((*it)->getUhOld()) {
	    if(D2UhOldIq) {
	      (*it)->evalSecondOrder(numPoints, 
				     uhOldIq, grdUhOldIq, D2UhOldIq, 
				     result, factor);
	    }
	    if(grdUhOldIq) {
	      (*it)->evalFirstOrderGrdPsi(numPoints, 
					  uhOldIq, grdUhOldIq, D2UhOldIq, 
					  result, -factor);
	      (*it)->evalFirstOrderGrdPhi(numPoints, 
					  uhOldIq, grdUhOldIq, D2UhOldIq, 
					  result, -factor);
	    }
	    if(uhOldIq) {
	      (*it)->evalZeroOrder(numPoints, 
				   uhOldIq, grdUhOldIq, D2UhOldIq, 
				   result, -factor);
	    }
	  } else {
	    int iq;
	    double *fx = GET_MEMORY(double, numPoints);
	    for(iq = 0; iq < numPoints; iq++) {
	      fx[iq] = 0.0;
	    }
	    (*it)->getC(elInfo, numPoints, fx);

	    for(iq = 0; iq < numPoints; iq++) {
	      result[iq] -= factor * fx[iq];
	    }
	    FREE_MEMORY(fx, double, numPoints);
	  }
	}
      }
  }

  double Estimator::estimate(double ts)
  {
    FUNCNAME("Estimator::estimate()");
  
    init(ts);
  
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while(elInfo) {
      estimateElement(elInfo);
      elInfo = stack.traverseNext(elInfo);
    }

    exit();
    return est_sum;
  }

  void ResidualEstimator::init(double ts)
  {
    FUNCNAME("ResidualEstimator::init()");

    timestep = ts;

    mesh = uh[row == -1 ? 0 : row]->getFESpace()->getMesh();

    numSystems = static_cast<int>(uh.size());
    TEST_EXIT(numSystems > 0)("no system set\n");

    dim = mesh->getDim();
    basFcts = GET_MEMORY(const BasisFunction*, numSystems);
    quadFast = GET_MEMORY(FastQuadrature*, numSystems);

    degree = 0;
    for(system = 0; system < numSystems; system++) {
      basFcts[system] = uh[system]->getFESpace()->getBasisFcts();
      degree = ::std::max(degree, basFcts[system]->getDegree());
    }

    degree *= 2;

    quad = Quadrature::provideQuadrature(dim, degree);
    numPoints = quad->getNumPoints();

    Flag flag = INIT_PHI | INIT_GRD_PHI;
    if(degree > 2) flag |= INIT_D2_PHI;

    for(system = 0; system < numSystems; system++) {
      quadFast[system] = FastQuadrature::provideFastQuadrature(basFcts[system], 
							       *quad, 
							       flag);
    }
  


    uhEl = GET_MEMORY(double*, numSystems);
    uhOldEl = timestep ? GET_MEMORY(double*, numSystems) : NULL;

    for(system = 0; system < numSystems; system++) {
      uhEl[system] = GET_MEMORY(double, basFcts[system]->getNumber()); 
      if(timestep)
	uhOldEl[system] = GET_MEMORY(double, basFcts[system]->getNumber());
    }

    //const double *uhQP = NULL;
    uhQP = timestep ? GET_MEMORY(double, quad->getNumPoints()) : NULL;
    uhOldQP = 
      timestep ? GET_MEMORY(double, quad->getNumPoints()) : NULL;


    riq = GET_MEMORY(double, numPoints);

    TraverseStack stack;
    ElInfo *elInfo = NULL;

    // clear error indicators and mark elements for jumpRes
    elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL);
    while(elInfo) {
      elInfo->getElement()->setEstimation(0.0, row);
      elInfo->getElement()->setMark(1);
      //    elInfo->getElement()->setEstimation(0.0);
      elInfo = stack.traverseNext(elInfo);
    }

    est_sum = est_max = est_t_sum = est_t_max = 0.0;
    traverseFlag = 
      Mesh::FILL_NEIGH      |
      Mesh::FILL_COORDS     |
      Mesh::FILL_OPP_COORDS |
      Mesh::FILL_BOUND      |
      Mesh::FILL_GRD_LAMBDA |
      Mesh::FILL_DET        |
      Mesh::CALL_LEAF_EL;
  }

  void ResidualEstimator::exit()
  {
    FUNCNAME("ResidualEstimator::exit()");

    est_sum   = sqrt(est_sum);
    est_t_sum = sqrt(est_t_sum);

    for(system = 0; system < numSystems; system++) {
      FREE_MEMORY(uhEl[system], double, basFcts[system]->getNumber());
      if(timestep)
	FREE_MEMORY(uhOldEl[system], double, basFcts[system]->getNumber());    
    }

    FREE_MEMORY(uhEl, double*, numSystems);
    if(timestep)
      FREE_MEMORY(uhOldEl, double*, numSystems);

    if(timestep) {
      FREE_MEMORY(uhQP, double, numPoints);
      FREE_MEMORY(uhOldQP, double, numPoints);
    }

    MSG("estimate   = %.8e\n", est_sum);
    if(C3)
      MSG("time estimate   = %.8e\n", est_t_sum);

    FREE_MEMORY(riq, double, numPoints);
    FREE_MEMORY(basFcts, const BasisFunction*, numSystems);
    FREE_MEMORY(quadFast, FastQuadrature*, numSystems);
  }

  void ResidualEstimator::estimateElement(ElInfo *elInfo)
  {
    FUNCNAME("ResidualEstimator::estimateElement()");

    double est_el;
    double val;
    Element *el, *neigh;
    int iq;

    TEST_EXIT(numSystems > 0)("no system set\n");

    ::std::vector<Operator*>::iterator it;

    const WorldVector<double> *grdUh_qp = NULL;
    const WorldMatrix<double> *D2uhqp = NULL; 

    el = elInfo->getElement();

    double det = elInfo->getDet();
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();

    est_el = el->getEstimation(row);

    double h2 = h2_from_det(det, dim);

    for(iq = 0; iq < numPoints; iq++) {
      riq[iq] = 0.0;
    }

    for(system = 0; system < numSystems; system++) {

      if(matrix[system] == NULL) continue;

      // init assemblers
      ::std::vector<Operator*>::iterator it;

      for(it = const_cast<DOFMatrix*>(matrix[system])->getOperatorsBegin();
	  it != const_cast<DOFMatrix*>(matrix[system])->getOperatorsEnd(); 
	  ++it) 
	{
	  (*it)->getAssembler()->initElement(elInfo, quad);
	}

      for(it = const_cast<DOFVector<double>*>(fh[system])->getOperatorsBegin();
	  it != const_cast<DOFVector<double>*>(fh[system])->getOperatorsEnd(); 
	  ++it) 
	{
	  (*it)->getAssembler()->initElement(elInfo, quad);
	}
    
      //uh[system]->getLocalVector(el, uhEl);
      if(timestep) {
	TEST_EXIT(uhOld[system])("no uhOld\n");
	uhOld[system]->getLocalVector(el, uhOldEl[system]);
      }
      
      // ===== time and element residuals       
      if (timestep && (C0 || C3)) {
	//quadFast->uhAtQp(uhEl, uhQP);
	uh[system]->getVecAtQPs(elInfo, NULL, quadFast[system], uhQP);
	//quadFast->uhAtQp(uhOldEl, uhOldQP);
	uhOld[system]->getVecAtQPs(elInfo, NULL, quadFast[system], uhOldQP);
      
	if(C3 && uhOldQP && system == ::std::max(row, 0)) {
	  for (val = iq = 0; iq < numPoints; iq++) {
	    double tiq = (uhQP[iq] - uhOldQP[iq]);
	    val += quad->getWeight(iq)*tiq*tiq;
	  }
	  double v = C3*det*val;
	  est_t_sum += v;
	  est_t_max = max(est_t_max, v);
	}
      }

      if (C0) {
	for(it = const_cast<DOFMatrix*>(matrix[system])->getOperatorsBegin(); 
	    it != const_cast<DOFMatrix*>(matrix[system])->getOperatorsEnd(); 
	    ++it) 
	  {
	    if((*it)->zeroOrderTerms() && !uhQP) {
	      //uhQP = const_cast<double*>
	      //  (quadFast->uhAtQp(uhEl, static_cast<double*>(NULL)));
	      uhQP = const_cast<double*>
		(uh[system]->getVecAtQPs(elInfo, NULL, quadFast[system], NULL));
	    }
	    if((*it)->firstOrderTermsGrdPsi() || (*it)->firstOrderTermsGrdPhi() 
	       && !grdUh_qp) 
	      {
		grdUh_qp = 
		  //quadFast->grdUhAtQp(Lambda, 
		  //			  const_cast<const double*>(uhEl), 
		  //		  static_cast<WorldVector<double>*>(NULL));
		  uh[system]->getGrdAtQPs(elInfo, NULL, quadFast[system], NULL);
	      }
	    if((*it)->secondOrderTerms() && !D2uhqp) {
	      if(degree > 2)
		D2uhqp = 
		  // 		quadFast->D2UhAtQp(Lambda, 
		  // 				   uhEl, 
		  // 				   static_cast<WorldMatrix<double>*>(NULL));
		  uh[system]->getD2AtQPs(elInfo, NULL, quadFast[system], NULL);
	    }
	  }
	
	r(elInfo,
	  numPoints, 
	  uhQP,
	  grdUh_qp,
	  D2uhqp,
	  uhOldQP,
	  NULL,  // grdUhOldQP 
	  NULL,  // D2UhOldQP
	  matrix[system], 
	  fh[system],
	  quad,
	  riq);	
      }
    }

    // add integral over r square
    for(val = iq = 0; iq < numPoints; iq++) {
      val += quad->getWeight(iq)*riq[iq]*riq[iq];
    }
  
    if (timestep != 0.0 || norm == NO_NORM || norm == L2_NORM) 
      val = C0*h2*h2*det*val;
    else
      val = C0*h2*det*val;

    est_el += val;

    // ===== jump residuals 
    int neighbours = Global::getGeo(NEIGH, dim);
    int face;
      
    if(dim > 1) {
      Quadrature *surfaceQuad = Quadrature::provideQuadrature(dim-1, degree);
    
      int numPointsSurface = surfaceQuad->getNumPoints();

      Vector<WorldVector<double> > jump(numPointsSurface);
    
      if (C1) {
	for (face = 0; face < neighbours; face++) {
	  if ((neigh = const_cast<Element*>(elInfo->getNeighbour(face)))) {
	    if (neigh->getMark()) {
	      ElInfo *neighInfo = mesh->createNewElInfo();
	      WorldVector<int> faceIndEl, faceIndNeigh;
	      Element *neigh = 
		const_cast<Element*>(elInfo->getNeighbour(face));
	      int oppV = elInfo->getOppVertex(face);
	      DimVec<WorldVector<double> > LambdaNeigh(dim, NO_INIT);
	      double detNeigh;
	      DimVec<double> lambda(dim, NO_INIT);
	    
	      el->sortFaceIndices(face, &faceIndEl);
	      neigh->sortFaceIndices(oppV, &faceIndNeigh);
	    
	      neighInfo->setElement(const_cast<Element*>(neigh));
	      neighInfo->setFillFlag(Mesh::FILL_COORDS);
	    
	      int dow = Global::getGeo(WORLD);
		
	      int i, j, i1, i2;

	      for (i = 0; i < dow; i++)
		neighInfo->getCoord(oppV)[i] = elInfo->getOppCoord(face)[i];
		
	      // periodic leaf data ?
	      ElementData *ldp = el->getElementData()->getElementData(PERIODIC);

	      bool periodicCoords = false;

	      if(ldp) {
		::std::list<LeafDataPeriodic::PeriodicInfo>::iterator it;
		for(it = dynamic_cast<LeafDataPeriodic*>(ldp)->getInfoList().begin();
		    it != dynamic_cast<LeafDataPeriodic*>(ldp)->getInfoList().end();
		    ++it) 
		  {
		    if(it->elementSide == face) {
		      for(i=0; i < dim; i++) {
			i1 = faceIndEl[i];
			i2 = faceIndNeigh[i];

			for(j=0; j < dim; j++) {
			  if(i1 == el->getVertexOfPosition(INDEX_OF_DIM(dim-1, 
									dim),
							   face,
							   j))
			    {
			      break;
			    }
			}

			TEST_EXIT(j != dim)("vertex i1 not on face ???\n");

			neighInfo->getCoord(i2) = (*(it->periodicCoords))[j];
		      }
		      periodicCoords = true;
		      break;
		    }
		  }
	      }

	      if(!periodicCoords) {
		for (i = 0; i < dim; i++) {
		  i1 = faceIndEl[i];
		  i2 = faceIndNeigh[i];
		  for (j = 0; j < dow; j++)
		    neighInfo->getCoord(i2)[j] = elInfo->getCoord(i1)[j];
		}
	      }
		
	      Parametric *parametric = mesh->getParametric();
	      if(parametric) {
		neighInfo = parametric->addParametricInfo(neighInfo);
	      }
	      
	      detNeigh = abs(neighInfo->calcGrdLambda(LambdaNeigh));

	      Vector<WorldVector<double> > grdUhEl(numPointsSurface);
	      Vector<WorldVector<double> > grdUhNeigh(numPointsSurface);

	      for(iq = 0; iq < numPointsSurface; iq++) {
		jump[iq].set(0.0);
	      }

	      for(system = 0; system < numSystems; system++) {
		if(matrix[system] == NULL) continue;
	      
		uh[system]->getLocalVector(el, uhEl[system]);

		const double* uhNeigh;
		
		uhNeigh = uh[system]->getLocalVector(neigh, NULL);

		for (iq = 0; iq < numPointsSurface; iq++) {
		  lambda[face] = 0.0;
		  for (i = 0; i < dim; i++)
		    lambda[faceIndEl[i]] = surfaceQuad->getLambda(iq,i);
		  basFcts[system]->evalGrdUh(lambda, Lambda, uhEl[system], &grdUhEl[iq]);
		
		  lambda[oppV] = 0.0;
		  for (i = 0; i < dim; i++)
		    lambda[faceIndNeigh[i]] = surfaceQuad->getLambda(iq,i);
		  basFcts[system]->evalGrdUh(lambda, 
					     LambdaNeigh, 
					     uhNeigh, 
					     &grdUhNeigh[iq]);

		  grdUhEl[iq] -= grdUhNeigh[iq];
		}

		::std::vector<double*>::iterator fac;

		for(it = const_cast<DOFMatrix*>(matrix[system])->getOperatorsBegin(),
		      fac = const_cast<DOFMatrix*>(matrix[system])->getOperatorEstFactorBegin(); 
		    it != const_cast<DOFMatrix*>(matrix[system])->getOperatorsEnd(); 
		    ++it, ++fac) 
		  {
		    int i;
		    Vector<WorldVector<double> > localJump(numPointsSurface);
		    for(iq = 0; iq < numPointsSurface; iq++) {
		      localJump[iq].set(0.0);
		    }

		    (*it)->weakEvalSecondOrder(numPointsSurface,
					       grdUhEl.getValArray(),
					       localJump.getValArray());
		    double factor = *fac ? **fac : 1.0;
		    if(factor != 1.0) {
		      for(i = 0; i < numPointsSurface; i++) {
			localJump[i] *= factor;
		      }
		    }

		    for(i = 0; i < numPointsSurface; i++) {
		      jump[i] += localJump[i];
		    }
		  }
	      }

	      for(val =iq = 0; iq < numPointsSurface; iq++) {
		val += surfaceQuad->getWeight(iq)*(jump[iq]*jump[iq]);
	      }
	      
	      double d =  0.5*(det + detNeigh);

	      if (norm == NO_NORM || norm == L2_NORM)
		val *= C1*h2_from_det(d, dim) * d;
	      else
		val *= C1*d;
	      
	      if(parametric) {
		neighInfo = parametric->removeParametricInfo(neighInfo);
	      }

	      DELETE neighInfo;

	      neigh->setEstimation(neigh->getEstimation(row) + val, row);

	      est_el += val;
	    }
	  }
	}
	
	val = fh[::std::max(row, 0)]->
	  getBoundaryManager()->
	  boundResidual(elInfo, matrix[::std::max(row, 0)], uh[::std::max(row, 0)]);
	if (norm == NO_NORM || norm == L2_NORM)
	  val *= C1 * h2;
	else
	  val *= C1;
	
	est_el += val;
      }
    }
  
    el->setEstimation(est_el, row);

    est_sum += est_el;
    est_max = max(est_max, est_el);

    elInfo->getElement()->setMark(0);
  }

  Estimator::Estimator(::std::string name_, int r) 
    : name(name_),
      norm(NO_NORM),
      row(r)
  {
    GET_PARAMETER(0, name + "->error norm", "%d", &norm);
  }

  ResidualEstimator::ResidualEstimator(::std::string name, int r) 
    : Estimator(name, r),
      C0(1.0), 
      C1(1.0), 
      C2(1.0), 
      C3(1.0)
  {
    GET_PARAMETER(0, name + "->C0", "%f", &C0);
    GET_PARAMETER(0, name + "->C1", "%f", &C1);
    GET_PARAMETER(0, name + "->C2", "%f", &C2);
    GET_PARAMETER(0, name + "->C3", "%f", &C3);

    C0    = C0 > 1.e-25 ? sqr(C0) : 0.0;
    C1    = C1 > 1.e-25 ? sqr(C1) : 0.0;
    C2    = C2 > 1.e-25 ? sqr(C2) : 0.0;
    C3    = C3 > 1.e-25 ? sqr(C3) : 0.0;
  }

}
