#include <vector>
#include "ILUTPreconditioner.h"
#include "DOFVector.h"
#include "DOFMatrix.h"
#include "DOFIterator.h"
#include "MatVecMultiplier.h"

namespace AMDiS {

  void ILUTPreconditioner::init()
  {
    FUNCNAME("ILUTPreconditioner::init()");

    TEST_EXIT(matrix[row])("no matrix\n");

    clock_t first = clock();

    // Make a copy of the system matrix.
    ilu_ = (*(*matrix[row]));

    int matrixSize = ilu_.getRowFESpace()->getAdmin()->getUsedSize();

    diag_.resize(matrixSize);

    // For i = 1, ..., n, Do
    for (int i = 0; i < matrixSize; i++) {
      // We do not make a real copy of the row, but instead w is a pointer to
      // the current row.
      ::std::vector<MatEntry> *w = &(ilu_[i]);

      
      // Calculate tau_i by calculating the 2-norm of the current row and
      // multiplying the result with the parameter tau.
      double taui = 0.0;
      for (int col1 = 0; col1 < static_cast<int>((*w).size()); col1++) {
	taui += (*w)[col1].entry * (*w)[col1].entry;
      }
      taui = sqrt(taui) * tau_;

      
      // For k = 1, ..., i - 1 and when w_k not= 0, Do
      for (int col1 = 0; col1 < static_cast<int>((*w).size()); col1++) {
	// Pointer to the w_ik element
	MatEntry *w_k = &((*w)[col1]);

	// Get the logical index k of the current element of row w.
	int k = (*w_k).col;

	if ((k < i) &&                   // must be smaller than i due to the loop condition
	    (k >= 0) &&                  // must be a valid entry
	    ((*w_k).entry != 0.0)) {     // must be non zero due to the loop condition

	  // w_k := w_k / a_kk   (note that the diagonal elements are always stored in the first entry of a row)
	  (*w_k).entry /= (*(*matrix[row]))[k][0].entry;

	  // Dropping rule to w_k
	  // It is checked if the absolute value of w_k is smaller than tau_i
	  if (fabs((*w_k).entry) < taui) {
	    // Do not delete the element yet but set it to zero, so it will be deleted
	    // when the second dropping rule will be applied.
	    (*w_k).entry = 0.0;
	  } else {
	    // w := w - w_k * u_k*
	    // To compute the result, all elements of u_k are multiplied with w_k. Then
	    // we check if there is an element in w with the same column index. If yes,
	    // the result of the multiplication is substracted from the corresponding 
	    // element in w. Otherwise, a new element in w with the negative value of
	    // the multiplication is created.
	    for (int col2 = 0; col2 < static_cast<int>(ilu_[k].size()); col2++) {

	      // Pointer to the current element of row u_k*
	      MatEntry *u_kj = &ilu_[k][col2];

	      // We are in the k-th row of matrix ilu_. The row is devided into L and
	      // U factorization. Hence, all elements of U in ilu_ must have a column
	      // index that is greater than k.
	      if (((*u_kj).col > k) &&               // must be a U element
		  ((*u_kj).entry != 0.0)) {          // must be non zero

		// w_k * u_kj
		double v = (*w_k).entry * (*u_kj).entry;

		// search for the physical index of the j-th column in row w.
		int index = -1;
		for (int z = 0; z < static_cast<int>((*w).size()); z++) {
		  if ((*w)[z].col == (*u_kj).col) {
		    index = z;
		    break;
		  }
		}

		if (index >= 0) {
		  // If there is a j-th column in row w, make the substraction.
		  (*w)[index].entry -= v;
		} else {
		  // Otherwise, if the absolut value of the result of the multiplication
		  // is greater than the relative tolerance, create a new element in w.
		  if (fabs(v) > taui) {
		    MatEntry newEntry = {(*u_kj).col, -v};
		    (*w).push_back(newEntry);
		  }
		}
	      }
      
	    }
	  }
	}
      }

      // Apply second dropping rule.
      // First, remove all elements which absolut value is smaller than tau_i.
      (*w).erase(remove_if((*w).begin(), (*w).end(), MatEntryValueLess(taui)),
		 (*w).end());

      // Calculate the number of elements in the L-part of the current row.
      int nL = 0;
      for (int k = 1; k < static_cast<int>((*w).size()); k++) {
	if ((*w)[k].col < i) {
	  nL++;
	}
      }
      // Calculate the number of elements in the U-part of the current row.
      int nU = (*w).size() - nL - 1;

      // Now we check, if there are more elements in L- and U-parts of the row
      // than we should store. If it is so, the final number of elements is
      // reduced to p.
      if (nL > p_) {
	nL = p_;
      }
      if (nU > p_) {
	nU = p_;
      }

      // First, we sort the row in such a way that the first element will stay to be
      // the diagonal element, then all the elements of the L-part follows and after 
      // them all the elements of the U-part. Because we are in the i-th row, elements
      // of L must have a column index smaller than i.
      // itMiddle is then an iterator to the first element of the U-part.
      ::std::vector<MatEntry>::iterator itMiddle = 
	  partition((*w).begin() + 1, (*w).end(), MatEntryColLess(i));

      // If there are elements in the L-part:
      if (nL > 0) {
	// Search for the nL largest elements in the L-part of the row.
	nth_element((*w).begin() + 1, 
		    (*w).begin() + 1 + nL, 
		    itMiddle - 1, 
		    CmpMatEntryAbsValueGreater());

	// Delete all the elements that do not belong to the nL largest elements
	// in the L-part.
	itMiddle = (*w).erase((*w).begin() + nL + 1, itMiddle);
      }

      // If there are elements in the U-part:
      if (nU > 0) {
	// Search for the nU largest elements in the U-part of the row.
	nth_element(itMiddle,
		    itMiddle + nU,
		    (*w).end(),
		    CmpMatEntryAbsValueGreater());

	// Delete all the elements that do not belong to the nU largest elements
	// in the L-part.
	(*w).erase(itMiddle + nU, (*w).end());
      }
     
      // Store the diagonal element and sort the line.
      diag_[i] = (*w)[0].entry;
      sort((*w).begin(), (*w).end(), CmpMatEntryCol());
    }


    MSG("preconditioning of the coefficient matrix needed %.5f seconds\n",
	TIME_USED(first,clock()));
  }


  void ILUTPreconditioner::exit()
  {
    ilu_.clear();
  }

  void ILUTPreconditioner::precon(DOFVector<double>* x)
  {    
    // The solution of P * x = b must be calculated in the
    // following. Note that the vector b is stored in x to
    // saving space.
    // We solve the equation with forward and backward substitution
    // of the factorization P = L * U.
 
    DOFVector<double>::Iterator itx(x, USED_DOFS);

    // Forward substitution: L * y = b
    // We use x to save the new results of y. Note that all diagonal 
    // elements of L are 1, hence we do not need to devide the term
    // by the diagonal elements.
    int i = 0;
    for (itx.reset(); !itx.end(); ++itx, ++i)
      {
	// y_i = b_i
	// Must not be assigned directly, since x = b and y will take
	// the results of y.

	// y_i = b_i - \sum_{1}^{i-1} ( l_{i,k} * y_k )
	for (int col = 0; col < static_cast<int>(ilu_[i].size()); col++) {
	  int k = ilu_[i][col].col;
	  if (k >= 0) {
	    if (k < i) {
	      (*itx) -= ilu_[i][col].entry * (*x)[k];
	    } else {
	      // If for an entry k < i does not hold, it will be also false
	      // for all other elements in th column, because the columns 
	      // are sorted.
	      break;
	    }
	  }
	}
      }
    
    
    // Backward substitution: R * x = y
    do {
      itx--;
      i--;
      
      // x_i = y_i - \sum_{i + 1}^(n} ( r_{i,k} * x_k )
      // The sum is build backwards, since the directon of calculation does not 
      // matter for the result, but we can break the search if k > i does not 
      // hold (the cols of ilu_ are sorted!) and so make the calculation faster.
      for (int col = static_cast<int>(ilu_[i].size()) - 1; col >= 0; col--) {
 	int k = ilu_[i][col].col;
  	if (k > i) {
  	  (*itx) -= ilu_[i][col].entry * (*x)[k];
  	} else {
	  break;
	}
      }

      // x_i = x_i / r_{i,i}
      (*itx) /= diag_[i];
      
    } while (!itx.begin());
  }

}
