#include "LeafData.h"
#include "Element.h"
#include "Mesh.h"

namespace AMDiS {

bool LeafDataEstimatable::refineElementData(Element* parent, 
					    Element* child1,
					    Element* child2,
					    int elType)
 
{
  ElementData::refineElementData(parent, child1, child2, elType);

  LeafDataEstimatable *child1Data = 
    NEW LeafDataEstimatable(child1->getElementData());
  LeafDataEstimatable *child2Data = 
    NEW LeafDataEstimatable(child2->getElementData());

  child1Data->setErrorEstimate(0, errorEstimate / 2.0);
  child2Data->setErrorEstimate(0, errorEstimate / 2.0);

  child1->setElementData(child1Data);
  child2->setElementData(child2Data);

  return true;
}

void LeafDataEstimatable::coarsenElementData(Element* parent, 
					     Element* thisChild,
					     Element* otherChild,
					     int elTypeParent)
 
{
  TEST_EXIT(otherChild->deleteElementData(ESTIMATABLE))
    ("couldn't delete LeafDataEstimatable at otherChild");
  parent->setElementData(NEW LeafDataEstimatable(parent->getElementData()));
  ElementData::coarsenElementData(parent, thisChild, otherChild, elTypeParent);
}

bool LeafDataEstimatableVec::refineElementData(Element* parent, 
					       Element* child1,
					       Element* child2,
					       int elType)
 
{
  ElementData::refineElementData(parent, child1, child2, elType);
  //int size = dynamic_cast<int>(errorEstimate.size());
  child1->setElementData(NEW LeafDataEstimatableVec(//size,
						    child1->getElementData()));
  child2->setElementData(NEW LeafDataEstimatableVec(//size,
						    child2->getElementData()));

  return true;
}

void LeafDataEstimatableVec::coarsenElementData(Element* parent, 
						Element* thisChild,
						Element* otherChild,
						int elTypeParent)
 
{
  TEST_EXIT(otherChild->deleteElementData(ESTIMATABLE))
    ("couldn't delete LeafDataEstimatableVec at otherChild");
  //int size = errorEstimate.getSize();
  parent->setElementData(NEW LeafDataEstimatableVec(//size,
						    parent->getElementData()));
  ElementData::coarsenElementData(parent, thisChild, otherChild, elTypeParent);
}

bool LeafDataCoarsenable::refineElementData(Element* parent, 
					    Element* child1,
					    Element* child2,
					    int elType)
 {
  ElementData::refineElementData(parent, child1, child2, elType);
  child1->setElementData(NEW LeafDataCoarsenable(child1->getElementData()));
  child2->setElementData(NEW LeafDataCoarsenable(child2->getElementData()));
  return true;
}

void LeafDataCoarsenable::coarsenElementData(Element* parent, 
					     Element* thisChild,
					     Element* otherChild,
					     int elTypeParent)
{
  TEST_EXIT(otherChild->deleteElementData(COARSENABLE))
    ("couldn't delete LeafDataCoarsenable at otherChild");
  parent->setElementData(NEW LeafDataCoarsenable(parent->getElementData()));
  ElementData::coarsenElementData(parent, thisChild, otherChild, elTypeParent);
}

bool LeafDataCoarsenableVec::refineElementData(Element* parent, 
					       Element* child1,
					       Element* child2,
					       int elType)
{
  ElementData::refineElementData(parent, child1, child2, elType);
  child1->setElementData(NEW LeafDataCoarsenableVec(child1->getElementData()));
  child2->setElementData(NEW LeafDataCoarsenableVec(child2->getElementData()));
  return true;
}

void LeafDataCoarsenableVec::coarsenElementData(Element* parent, 
						Element* thisChild,
						Element* otherChild,
						int elTypeParent) 
{
  TEST_EXIT(otherChild->deleteElementData(COARSENABLE))
    ("couldn't delete LeafDataCoarsenableVec at otherChild");
  parent->setElementData(NEW LeafDataCoarsenableVec(parent->getElementData()));
  ElementData::coarsenElementData(parent, thisChild, otherChild, elTypeParent);
}

bool LeafDataPeriodic::refineElementData(Element* parent, 
					 Element* child1,
					 Element* child2,
					 int elType) 
{
  ElementData::refineElementData(parent, child1, child2, elType);

  int i, j;

  Element* child[2] = {child1, child2};

  int dim = parent->getMesh()->getDim();

  LeafDataPeriodic *ld[2] = {NULL, NULL};

  ::std::list<LeafDataPeriodic::PeriodicInfo>::iterator it;

  for(it = periodicInfoList.begin();
      it != periodicInfoList.end();
      ++it)
  {
    BoundaryType type = it->type;

    int parentSide = it->elementSide;
    
    int mode = it->periodicMode;

    // for both children
    for(i=0; i < 2; i++) {

      // get childs side
      int sideOfChild = parent->getSideOfChild(i, parentSide, elType);
      if(sideOfChild == -1) continue;

      // create new leaf data if necessary
      if(!ld[i]) {
	ld[i] = NEW LeafDataPeriodic(child[i]->getElementData());
	//ld[i]->newDOF = newDOF;
	child[i]->setElementData(ld[i]);
      }
    
      // create new periodic coords
      DimVec<WorldVector<double> > coords(dim-1, NO_INIT);

      // create new periodic dofs
      //DimVec<DegreeOfFreedom> dofs(dim-1, DEFAULT_VALUE, -1);

      // for each vertex of childs side
      for(j=0; j < dim; j++) {
	// get parents vertex nr.
	int childVertex = child[i]->getVertexOfPosition(INDEX_OF_DIM(dim-1,dim),
							sideOfChild,
							j);
	int parentVertex = child[i]->getVertexOfParent(i, childVertex, elType);
	
	if(parentVertex == -1) {
	  // create coords for new vertex
	  WorldVector<double> newCoords;
	  newCoords = (*(it->periodicCoords))[0];
	  newCoords += (*(it->periodicCoords))[1];
	  newCoords *= 0.5;
	  coords[j] = newCoords;
	  //dofs[j] = newDOF;
	  //TEST_EXIT(newDOF != -1)("newDOF not set\n");
	} else {
	  int posAtSide = parent->getPositionOfVertex(parentSide, 
						      parentVertex);
	  coords[j] = (*(it->periodicCoords))[posAtSide];
	  //dofs[j] = it->periodicDOFs[posAtSide];
	}
      }

      ld[i]->addPeriodicInfo(mode, type, sideOfChild, &coords);

//       MSG("parent %d, child%d %d, periodic dofs %d %d\n",
// 	  parent->getIndex(),
// 	  i,
// 	  child[i]->getIndex(),
// 	  dofs[0], dofs[1]);
    }
  }

  return false;
}

LeafDataPeriodic::PeriodicInfo::PeriodicInfo(const PeriodicInfo &rhs) 
{
  periodicMode = rhs.periodicMode;
  type = rhs.type;
  elementSide = rhs.elementSide;
  int i, dim;
  if(rhs.periodicCoords) {
    dim = rhs.periodicCoords->getSize() - 1;
    periodicCoords = NEW DimVec<WorldVector<double> >(dim, NO_INIT);
    for(i = 0; i < dim + 1; i++) {
      (*periodicCoords)[i] = (*(rhs.periodicCoords))[i];
    }
    
  } else {
    periodicCoords = NULL;
  }
}

LeafDataPeriodic::PeriodicInfo::PeriodicInfo(int mode,
					     BoundaryType t,
					     int side,
					     const DimVec<WorldVector<double> > *coords)
  : periodicMode(mode),
    type(t),
    elementSide(side),
    periodicCoords(NULL)
{
  int i, dim;
  if(coords) {
    dim = coords->getSize() - 1;
    periodicCoords = NEW DimVec<WorldVector<double> >(dim, NO_INIT);
    for(i = 0; i < dim + 1; i++) {
      (*periodicCoords)[i] = (*coords)[i];
    }
  }
}

// void LeafDataPeriodicRefinable::refineLeafData(Element* parent, 
// 					       Element* child1,
// 					       Element* child2) 
// {
//   if(decoratedLeafData) {
//     LeafDataPeriodic *ldp = 
//       (LeafDataPeriodic*) decoratedLeafData->getDecoratedLeafData(PERIODIC);
//     TEST_EXIT(ldp)("no periodic leaf data\n");

//     LeafDataRefinable *ldr = (LeafDataRefinable*)
//       decoratedLeafData->getDecoratedLeafData(REFINABLE);
//     if(ldr) {
//       ldr->refineLeafData(parent, child1, child2);
//     }

//     int dim = parent->getMesh()->getDim();

//     LeafDataPeriodic *ld1 = NULL;
//     LeafDataPeriodic *ld2 = NULL;

//     ::std::list<LeafDataPeriodic::PeriodicInfo>::iterator it;

//     switch(dim) {
//     case 1:  // ----- dim = 1 -----
//       // for each periodic info
//       for(it = ldp->periodicInfoList.begin();
// 	  it != ldp->periodicInfoList.end();
// 	  ++it)
//       {
// 	switch(it->elementSide) {
// 	case 0: // ----- side 0 -----
// 	  if(!ld1)
// 	    ld1 = NEW LeafDataPeriodic(child1->getLeafData());
// 	  ld1->addPeriodicInfo(0, it->periodicCoords);
// 	  child1->setLeafData(NEW LeafDataPeriodicRefinable
// 			      (NEW LeafDataPeriodicCoarsenable(ld1))
// 			     );
// 	  break;
// 	case 1: // ----- side 1 -----
// 	  if(!ld2)
// 	    ld2 = NEW LeafDataPeriodic(child2->getLeafData());
// 	  ld2->addPeriodicInfo(0, it->periodicCoords);
// 	  child2->setLeafData(NEW LeafDataPeriodicRefinable
// 			      (NEW LeafDataPeriodicCoarsenable(ld2))
// 			     );
// 	  break;
// 	default: ERROR_EXIT("invalid element side\n");
// 	}
//       }
//       break;
//     case 2:  // ----- dim = 2 -----
//       // for each periodic info
//       for(it = ldp->periodicInfoList.begin();
// 	  it != ldp->periodicInfoList.end();
// 	  ++it)
//       {
// 	switch(it->elementSide) {
// 	case 0: // ----- side 0 -----
// 	  if(!ld2)
// 	    ld2 = NEW LeafDataPeriodic(child2->getLeafData());
// 	  ld2->addPeriodicInfo(2, it->periodicCoords);
// 	  child2->setLeafData(NEW LeafDataPeriodicRefinable
// 			      (NEW LeafDataPeriodicCoarsenable(ld2))
// 			     );
// 	  break;
// 	case 1: // ----- side 1 -----
// 	  if(!ld1)
// 	    ld1 = NEW LeafDataPeriodic(child1->getLeafData());
// 	  ld1->addPeriodicInfo(2, it->periodicCoords);
// 	  child1->setLeafData(NEW LeafDataPeriodicRefinable
// 			      (NEW LeafDataPeriodicCoarsenable(ld1))
// 			     );
// 	  break;
// 	case 2: // ----- side 2 -----
// 	  {
// 	    if(!ld1)
// 	      ld1 = NEW LeafDataPeriodic(child1->getLeafData());
// 	    if(!ld2)
// 	      ld2 = NEW LeafDataPeriodic(child2->getLeafData());

// 	    DimVec<WorldVector<double> > coords1(it->periodicCoords);
// 	    DimVec<WorldVector<double> > coords2(it->periodicCoords);

// 	    WorldVector<double> newCoords;
// 	    newCoords = it->periodicCoords[0];
// 	    newCoords += it->periodicCoords[1];
// 	    newCoords *= 0.5;

// 	    coords1[1] = newCoords;
// 	    coords2[0] = newCoords;

// 	    ld1->addPeriodicInfo(0, coords1);
// 	    ld2->addPeriodicInfo(1, coords2);

// 	    child1->setLeafData(NEW LeafDataPeriodicRefinable
// 				(NEW LeafDataPeriodicCoarsenable(ld1))
// 				);
// 	    child2->setLeafData(NEW LeafDataPeriodicRefinable
// 				(NEW LeafDataPeriodicCoarsenable(ld2))
// 				);
// 	  }
// 	  break;
// 	default: ERROR_EXIT("invalid element side\n");
// 	}
//       }
//       break;
//     case 3: // ----- dim = 3 -----
//       ERROR_EXIT("not yet for 3d\n");
//       // for each periodic info
//       for(it = ldp->periodicInfoList.begin();
// 	  it != ldp->periodicInfoList.end();
// 	  ++it)
//       {
// 	switch(it->elementSide) {
// 	case 0: // ----- side 0 -----
// 	  break;
// 	case 1: // ----- side 1 -----
// 	  break;
// 	case 2: // ----- side 2 -----
// 	  break;
// 	case 3: // ----- side 3 -----
// 	  break;
// 	default: ERROR_EXIT("invalid element side\n");
// 	}
//       }
//       break;
//     default: ERROR_EXIT("invalid dim\n");
//     }
//   }
// }

// LeafData *LeafData::createLeafData(::std::string type) 
// {
//   if(type == "NULL") return NULL;
//   if(type == "LeafDataEstimatable") return new LeafDataEstimatable;
//   if(type == "LeafDataEstimatableVec") return new LeafDataEstimatableVec;
//   if(type == "LeafDataCoarsenable") return new LeafDataCoarsenable;
//   if(type == "LeafDataCoarsenableVec") return new LeafDataCoarsenableVec;
//   if(type == "LeafDataPeriodic") return new LeafDataPeriodic;

//   ERROR_EXIT("unknown type\n");
//   return NULL;
// }

}
