#include "PeriodicBC.h"
#include "BasisFunction.h"
#include "DOFVector.h"
#include "LeafData.h"
#include "DOFMatrix.h"
#include "Traverse.h"
#include "Boundary.h"
#include "VertexVector.h"

namespace AMDiS {

  ::std::vector<PeriodicDOFMapping*> PeriodicDOFMapping::mappings_;

  PeriodicDOFMapping* 
  PeriodicDOFMapping::providePeriodicDOFMapping(const BasisFunction *basFcts)
  {
    ::std::vector<PeriodicDOFMapping*>::iterator it;
    ::std::vector<PeriodicDOFMapping*>::iterator end = mappings_.end();
    for(it = mappings_.begin(); it != end; ++it) {
      if((*it)->basFcts_ == basFcts) {
	return *it;
      }
    }
    PeriodicDOFMapping *newMapping = NEW PeriodicDOFMapping(basFcts);
    mappings_.push_back(newMapping);
    return newMapping;
  }


  PeriodicDOFMapping::PeriodicDOFMapping(const BasisFunction *basFcts) 
    : basFcts_(basFcts) 
  {
    FUNCNAME("PeriodicDOFMapping::PeriodicDOFMapping()");
    TEST_EXIT(basFcts_->getDim() > 1)("dim == 1\n");  
    int i, num = basFcts_->getNumber();
    DimVec<double> *lambda;
    for(i = 0; i < num; i++) {
      lambda = basFcts_->getCoords(i);
      indexOfCoords_[*lambda] = i;
    }
  }

  PeriodicDOFMapping::~PeriodicDOFMapping()
  {
    ::std::map<DimVec<int>, DegreeOfFreedom*, DimVecLess<int> >::iterator it;
    for(it = dofPermutation_.begin(); it != dofPermutation_.end(); ++it) {
      if(it->second) {
	FREE_MEMORY(it->second, DegreeOfFreedom, basFcts_->getNumber());
      }
    }
  }

  const DegreeOfFreedom 
  *PeriodicDOFMapping::getDOFPermutation(const DimVec<int> &vertexPermutation)
  {
    FUNCNAME("PeriodicDOFMapping::getDOFPermutation()");

    if(dofPermutation_[vertexPermutation] == NULL) {
      //MSG("new dof permutation needed\n");

      int i, j;
      int dim = basFcts_->getDim();
      int num = basFcts_->getNumber();

      int sum = 0;
      for(i = 0; i < dim + 1; i++) {
	sum += i - vertexPermutation[i];
	TEST_EXIT(vertexPermutation[i] < dim + 1)
	  ("invalid vertexPermuation\n");
      }
      TEST_EXIT(sum == 0)("invalid vertexPermutation\n");

      // create dof permutation
      DimVec<double> *lambda;
      DimVec<double> newLambda(dim, NO_INIT);

      DegreeOfFreedom *mapping = GET_MEMORY(DegreeOfFreedom, num); 

      for(i = 0; i < num; i++) {
	lambda = basFcts_->getCoords(i);
	for(j = 0; j < dim + 1; j++) {
	  newLambda[vertexPermutation[j]] = (*lambda)[j];
	}
	mapping[i] = indexOfCoords_[newLambda];
      }

      dofPermutation_[vertexPermutation] = mapping;

      //     MSG("vertex permutation: ");
      //     for(i = 0; i < dim+1; i++) {
      //       MSG("%d ", vertexPermutation[i]);
      //     }
      //     MSG("\n");

      //     MSG("dof permutation: ");
      //     for(i = 0; i < num; i++) {
      //       MSG("%d ", dofPermutation_[vertexPermutation][i]);
      //     }
      //     MSG("\n");
      //     WAIT;
    }

    return dofPermutation_[vertexPermutation];
  }


  PeriodicBC::PeriodicBC(BoundaryType type, FiniteElemSpace *rowFESpace_) 
    : BoundaryCondition(type, rowFESpace_, NULL),
      masterMatrix_(NULL)
  {
    if(rowFESpace->getMesh()->getDim() > 1) {
      periodicDOFMapping_ = 
	PeriodicDOFMapping::providePeriodicDOFMapping(rowFESpace_->getBasisFcts());
    } else {
      periodicDOFMapping_ = NULL;
    }
  }

  PeriodicBC::~PeriodicBC()
  {}

  void PeriodicBC::initMatrix(DOFMatrix* matrix)
  {
    FUNCNAME("PeriodicBC::initMatrix()");

    //   MSG("begin initMatrix...\n");

    if(!masterMatrix_) {
      masterMatrix_ = matrix;

      Mesh *mesh = matrix->getRowFESpace()->getMesh();

      associated_ = mesh->getPeriodicAssociations()[boundaryType];
      
      TEST_EXIT(associated_)("no associations for periodic boundary condition %d\n",
			     boundaryType);

      const BasisFunction *basFcts = rowFESpace->getBasisFcts();
      int num = basFcts->getNumber();

      neighIndices_ = GET_MEMORY(DegreeOfFreedom, num);
    }

    //MSG("end initMatrix...\n");
  }

  void PeriodicBC::fillBoundaryCondition(DOFMatrix             *matrix,
					 ElInfo                *elInfo,
					 const DegreeOfFreedom *dofIndices,
					 const BoundaryType    *localBound,
					 int                    nBasFcts)
  {
    if(matrix == masterMatrix_) {
  
      int dim = rowFESpace->getMesh()->getDim();
      if(dim > 1) {
	int i, j;
	DOFAdmin *admin = rowFESpace->getAdmin();

	FixVec<int, WORLD> elFace(dim, NO_INIT);
	FixVec<int, WORLD> neighFace(dim, NO_INIT);
	DimVec<int> vertexPermutation(dim, NO_INIT);

	const BasisFunction *basFcts = rowFESpace->getBasisFcts();
	int num = basFcts->getNumber();

	Element *element = elInfo->getElement();

	DimVec<DegreeOfFreedom> periodicDOFs(dim-1, NO_INIT);

	int vertex, index, side;

	GeoIndex sideGeoIndex = INDEX_OF_DIM(dim-1, dim);

	for(side = 0; side < dim + 1; side++) {

	  if(elInfo->getBoundary(sideGeoIndex, side) == boundaryType) {

	    for(vertex = 0; vertex < dim; vertex++) {
	      index = element->getVertexOfPosition(sideGeoIndex,
						   side,
						   vertex);
	      periodicDOFs[vertex] = (*associated_)[dofIndices[index]];
	    }

	    Element *neigh = elInfo->getNeighbour(side);

	    basFcts->getLocalIndices(neigh, admin, neighIndices_);

	    int oppVertex = 0;
	    for(i = 0; i < dim + 1; i++) {
	      // get vertex permutation
	      if(i == side) {
		vertexPermutation[i] = 0;
	      } else {
		DegreeOfFreedom periodicDOF = 
		  periodicDOFs[element->getPositionOfVertex(side, i)];

		for(j = 0; j < dim + 1; j++) {
		  if(neigh->getDOF(j, 0) == periodicDOF) break;
		}
		vertexPermutation[i] = j;
	      }
	      oppVertex += i - vertexPermutation[i];
	    }
	    vertexPermutation[side] = oppVertex;

	    // get DOF permutation
	    const DegreeOfFreedom *dofPermutation = 
	      periodicDOFMapping_->getDOFPermutation(vertexPermutation);
	
	    // set associated dofs
	    for(i = 0; i < num; i++) {
	      if((*(basFcts->getCoords(i)))[side] == 0) {
		(*associated_)[dofIndices[i]] = neighIndices_[dofPermutation[i]];
	      }
	    }
	  }
	}
      }
    }
  }

  void PeriodicBC::exitMatrix(DOFMatrix* matrix)
  {
    FUNCNAME("PeriodicBC::exitMatrix()");

    TEST_EXIT(matrix)("no matrix\n");

    if(matrix == masterMatrix_) {
      const BasisFunction *basFcts = rowFESpace->getBasisFcts();
      int num = basFcts->getNumber();
      FREE_MEMORY(neighIndices_, DegreeOfFreedom, num);
      masterMatrix_ = NULL;
    }
    // ---------- different assemblage style --------

    //   DOFMatrix::Iterator matrixRow(matrix, USED_DOFS);
    //   for(matrixRow.reset(); !matrixRow.end(); ++matrixRow) {
    //     row = matrixRow.getDOFIndex();

    //     rowSize = matrixRow->size();
    //     newRow = ((*associated_)[row] == -1) ? row : (*associated_)[row];

    //     if(row < newRow) {
    //       for(i = 0; i < rowSize; i++) {
    // 	col = (*matrixRow)[i].col;
    // 	if(col == DOFMatrix::NO_MORE_ENTRIES) break;
    // 	if(col == DOFMatrix::UNUSED_ENTRY) continue;
    // 	newCol = ((*associated_)[col] == -1) ? col : (*associated_)[col];
    // 	if(col < newCol) {
    // 	  entry = (*matrixRow)[i].entry;
    // 	  (*matrixRow)[i].col = DOFMatrix::UNUSED_ENTRY;
    // 	  (*matrixRow)[i].entry = 0.0;
    // 	  matrix->addSparseDOFEntry(1.0, row, newCol, entry, true);
    // 	}
    //       }
    //       rowSize = matrixRow->size();
    //       for(i = 0; i < rowSize; i++) {
    // 	col = (*matrixRow)[i].col;
    // 	if(col == DOFMatrix::NO_MORE_ENTRIES) break;
    // 	if(col == DOFMatrix::UNUSED_ENTRY) continue;
    // 	entry = (*matrixRow)[i].entry;
    // 	matrix->addSparseDOFEntry(1.0, newRow, col, entry, true);
    //       }
    //       matrixRow->resize(2);
    //       (*matrixRow)[0].col = row;
    //       (*matrixRow)[0].entry = 100.0;1.0;
    //       (*matrixRow)[1].col = newRow;
    //       (*matrixRow)[1].entry = -100.0;-1.0;
    //     }
    //   }


    // --- method 2 ---

    //   for(matrixRow.reset(); !matrixRow.end(); ++matrixRow) {
    //     row = matrixRow.getDOFIndex();
    //     newRow = ((*associated_)[row] == -1) ? row : (*associated_)[row];
    //     if(row < newRow) {
    //       // remember old row matrix[row]
    //       ::std::vector<MatEntry> oldMatrixRow = *matrixRow;
    //       // add new row to old row
    //       ::std::vector<MatEntry> &newMatrixRow = matrix->getRow(newRow);
    //       rowSize = newMatrixRow.size();
    //       for(i = 0; i < rowSize; i++) {
    // 	col = newMatrixRow[i].col;
    // 	if(col == DOFMatrix::NO_MORE_ENTRIES) break;
    // 	if(col == DOFMatrix::UNUSED_ENTRY) continue;
    // 	newCol = ((*associated_)[col] == -1) ? col : (*associated_)[col];
    // 	entry = newMatrixRow[i].entry;
    // 	matrix->addSparseDOFEntry(1.0, row, newCol, entry, true);
    //       }
    //       // add old row to new row
    //       rowSize = oldMatrixRow.size();
    //       for(i = 0; i < rowSize; i++) {
    // 	col = oldMatrixRow[i].col;
    // 	if(col == DOFMatrix::NO_MORE_ENTRIES) break;
    // 	if(col == DOFMatrix::UNUSED_ENTRY) continue;
    // 	newCol = ((*associated_)[col] == -1) ? col : (*associated_)[col];
    // 	entry = oldMatrixRow[i].entry;
    // 	matrix->addSparseDOFEntry(1.0, newRow, newCol, entry, true);
    //       }      
    //     }
    //   }


    //   DOFVector<DegreeOfFreedom>::Iterator rowIt(associated_, USED_DOFS);
    //   DOFVector<DegreeOfFreedom>::Iterator colIt(associated_, USED_DOFS);


    //   Mesh *mesh = matrix->getRowFESpace()->getMesh();

    ::std::vector< ::std::vector<MatEntry> >::iterator rowIt;
    ::std::vector< ::std::vector<MatEntry> >::iterator rowEnd = matrix->end();
    int colIndex, rowSize;
    int row, col, newRow, newCol;
    double entry, *newEntryPtr;

    for(rowIt = matrix->begin(), row = 0; rowIt != rowEnd; ++rowIt, ++row) {
      rowSize = static_cast<int>(rowIt->size());
      newRow = (*associated_)[row];
      for(colIndex = 0; colIndex < rowSize; colIndex++) {
	col = (*rowIt)[colIndex].col;
	if(col == DOFMatrix::UNUSED_ENTRY) continue;
	if(col == DOFMatrix::NO_MORE_ENTRIES) break;
	newCol = (*associated_)[col];

	newEntryPtr = matrix->hasSparseDOFEntry(newRow, newCol);

	if((row < newRow) || 
	   ((row == newRow) && (col < newCol)) ||
	   !newEntryPtr) 
	  {
	    if(!newEntryPtr) {
	      entry = 0.5 * (*rowIt)[colIndex].entry;
	      (*rowIt)[colIndex].entry = entry;
	      matrix->addSparseDOFEntry(1.0, newRow, newCol, entry, false);
	    } else {
	      entry = 0.5 * ((*rowIt)[colIndex].entry + *newEntryPtr);
	      (*rowIt)[colIndex].entry = *newEntryPtr = entry;
	    }
	  }
      }
    }

    //MSG("end exitMatrix...\n");
  }

  void PeriodicBC::exitVector(DOFVectorBase<double>* vector)
  {
    FUNCNAME("PeriodicBC::exitVector()");

    double entry;

    DegreeOfFreedom dof;
    DegreeOfFreedom newDOF;

    DOFIterator<double> vecIt(vector, USED_DOFS);

    Mesh *mesh = vector->getFESpace()->getMesh();
    VertexVector *associated_ = mesh->getPeriodicAssociations()[boundaryType];

    for(vecIt.reset(); !vecIt.end(); ++vecIt) {
      dof = vecIt.getDOFIndex();

      newDOF = (*associated_)[dof];

      if(dof < newDOF) {
	// ---------- different assemblage style --------
	//       (*vector)[newDOF] += (*vector)[dof];
	//       (*vector)[dof] = 0.0;


	// --- method 2 ----
	//       //entry = (*vector)[dof];
	//       (*vector)[dof] += (*vector)[newDOF];
	//       (*vector)[newDOF] = (*vector)[dof];

	entry = ((*vector)[dof] + (*vector)[newDOF]) * 0.5;
	(*vector)[dof] = entry;
	(*vector)[newDOF] = entry;
      }
    }

    //MSG("end exitVector...\n");
  }

}
