#include <algorithm>

#include "ValueWriter.h"
#include "DOFVector.h"
#include "DOFAdmin.h"
#include "ElInfo.h"
#include "BasisFunction.h"

namespace AMDiS {

  const DOFAdmin* ValueWriter::admin = NULL;
  FILE* ValueWriter::valueFile = NULL;
  DOFVector<double>* ValueWriter::valueVec = NULL;
  DOFVector<int>* ValueWriter::interpPointInd = NULL;
  DOFVector< ::std::list<WorldVector<double> > >* ValueWriter::dofCoords = NULL;
  int ValueWriter::ni = 0;

  int ValueWriter::markFct(ElInfo *el_info) {
    int i, j, k;
    int node_offset, dof_offset, num_dofs, node, dof_index;
    const DegreeOfFreedom **dof = el_info->getElement()->getDOF();
    const Mesh *mesh = admin->getMesh();
    int dim = mesh->getDim();

    /* vertex dofs */
    dof_offset  = admin->getNumberOfPreDOFs(VERTEX);
  
    for (i = 0; i < mesh->getGeo(VERTEX); i++) {
      (*interpPointInd)[dof[i][dof_offset]] = -2; // mark as vertex

      // get coords of this vertex
      WorldVector<double> vertexCoords = el_info->getCoord(i);

      // search for coords at this dof
      ::std::list<WorldVector<double> >::iterator it =
	  find((*dofCoords)[dof[i][dof_offset]].begin(),
	       (*dofCoords)[dof[i][dof_offset]].end(),
	       vertexCoords);

      // coords not yet in list?
      if(it == (*dofCoords)[dof[i][dof_offset]].end()) {
	// add new coords to list
	(*dofCoords)[dof[i][dof_offset]].push_back(vertexCoords);
      }
    }
  
    for(i = 1; i <= dim; i++) {
      num_dofs = admin->getNumberOfDOFs(INDEX_OF_DIM(i, dim));
      node_offset = mesh->getNode(INDEX_OF_DIM(i, dim));
      dof_offset  = admin->getNumberOfPreDOFs(INDEX_OF_DIM(i, dim));
      for (j = 0; j < mesh->getGeo(INDEX_OF_DIM(i, dim)); j++) {
	node = node_offset + j;
	for(k = 0; k < num_dofs; k++) {
	  dof_index = dof_offset + k;
	  if ((*interpPointInd)[dof[node][dof_index]] == -1) {
	    (*interpPointInd)[dof[node][dof_index]] = -3; 
	    // mark as interp. point 
	    ni++;
	  }
	}
      }
    }

    return 0;
  }

  int ValueWriter::writeInterpolPoints(ElInfo *el_info)
  {
    int     i, j, k;
    int     node_offset, dof_offset, num_dofs, node, dof_index;
    const DegreeOfFreedom **dof = el_info->getElement()->getDOF();

    const Mesh *mesh = admin->getMesh();
    int dim = mesh->getDim();

    for(i = 1; i <= mesh->getDim(); i++) {
      num_dofs = admin->getNumberOfDOFs(INDEX_OF_DIM(i, dim));
      node_offset = mesh->getNode(INDEX_OF_DIM(i, dim));
      dof_offset  = admin->getNumberOfPreDOFs(INDEX_OF_DIM(i, dim));
      for (j = 0; j < mesh->getGeo(INDEX_OF_DIM(i, dim)); j++) {
	node = node_offset + j;
	for(k = 0; k < num_dofs; k++) {
	  dof_index = dof_offset + k;
	  fprintf(valueFile, "%d ", (*interpPointInd)[dof[node][dof_index]]);
	}
      }
    }

    fprintf(valueFile, "\n");
    return 0;
  }

  void ValueWriter::writeValuesOld(DOFVector<double> *values,
				   char*              fn,
				   double             time,
				   int                level,
				   Flag               traverseFlag)
  {
    FUNCNAME("ValueWriter::writeValues()");
    TEST_EXIT(values)("no values specified\n");

    ::std::string filename;
    if(fn) {
      filename.assign(fn);
    } else {
      filename.assign(values->getName());
    }

    const FiniteElemSpace *feSpace = values->getFESpace();
    Mesh *mesh  = feSpace->getMesh();

    interpPointInd = NEW DOFVector<int>(feSpace, "interpolation point indices");
    dofCoords = NEW DOFVector< ::std::list<WorldVector<double> > >(feSpace, "dof coords");

    admin = feSpace->getAdmin();

    valueVec = values;

    DOFVector<int>::Iterator intPointIt(interpPointInd, USED_DOFS);

    for(intPointIt.reset(); !intPointIt.end(); ++intPointIt) {
      (*intPointIt) = -1;
    }

    if (!(valueFile = fopen(filename.c_str(), "w")))
      {
	ERROR("could not open file %s for writing\n", filename.c_str());
      }

    ni = 0;
    mesh->traverse(level, traverseFlag | Mesh::FILL_COORDS, markFct);

    fprintf(valueFile, "mesh name: %s\n\n", 
	    values->getFESpace()->getMesh()->getName().c_str());
    fprintf(valueFile, "time: %e\n\n", time);
    fprintf(valueFile, "number of values: 1\n\n");
    fprintf(valueFile, "value description: %s\n", values->getName().c_str());
    fprintf(valueFile, "number of interpolation points: %d\n", ni);
    fprintf(valueFile, "type: scalar\n");
    fprintf(valueFile, "interpolation type: lagrange\n");
    fprintf(valueFile, "interpolation degree: %d\n",
	    feSpace->getBasisFcts()->getDegree());
    fprintf(valueFile, "end of description: %s\n\n", values->getName().c_str());


    /* ----- write vertex values -----*/
    DOFVector<double>::Iterator valueIt(valueVec, USED_DOFS);
    DOFVector< ::std::list<WorldVector<double> > >::Iterator coordIt(dofCoords, USED_DOFS);

    int i;

    fprintf(valueFile, "vertex values: %s\n", values->getName().c_str());

    for(intPointIt.reset(), valueIt.reset(), coordIt.reset();
	!intPointIt.end(); 
	++intPointIt, ++valueIt, ++coordIt) 
      {
	if(*intPointIt == -2) {
	  for(i=0; i < (int) coordIt->size(); i++) {
	    fprintf(valueFile, "%e\n", *valueIt);
	  }
	}
      }

    fprintf(valueFile, "\n\n");

    /* ----- write interpolation values ----- */
    fprintf(valueFile, "interpolation values: %s\n", values->getName().c_str());


    int nd = 0;
    for(intPointIt.reset(), valueIt.reset();
	!intPointIt.end(); 
	++intPointIt, ++valueIt) 
      {
	if(*intPointIt == -3) {
	  fprintf(valueFile, "%e\n", *valueIt);
	  *intPointIt = nd++;
	}
      }

    fprintf(valueFile, "\n\n");

    /* ----- write interpolation points for each simplex */
    fprintf(valueFile, "element interpolation points: %s\n", 
	    values->getName().c_str());
    mesh->traverse(level, traverseFlag,
		   writeInterpolPoints);
    fprintf(valueFile, "\n");
  
    fclose(valueFile);
    DELETE interpPointInd;
    DELETE dofCoords;
  }

  void ValueWriter::writeValues(DataCollector *dc,
				const ::std::string filename,
				double time,
				int level,
				Flag traverseFlag,
				bool (*writeElem)(ElInfo*))
  {
    FUNCNAME("ValueWriter::writeValues2");
    TEST_EXIT(dc)("no data collector\n");

    ::std::ofstream file;

    file.open(filename.c_str());

    file << "mesh name: " << dc->getMesh()->getName() << ::std::endl << ::std::endl;
    file << "time: " << ::std::scientific << time << ::std::endl << ::std::endl;
    file << "number of values: 1" << ::std::endl << ::std::endl;
    file << "value description: " << dc->getValues()->getName() << ::std::endl;
    file << "number of interpolation points: " << dc->getNumberInterpPoints()
	 << ::std::endl;
    file << "type: scalar" << ::std::endl;
    file << "interpolation type: lagrange" << ::std::endl;
    file << "interpolation degree: " << dc->getFeSpace()->getBasisFcts()->getDegree()
	 << ::std::endl;
    file << "end of description: " << dc->getValues()->getName()
	 << ::std::endl << ::std::endl;   

    /* ----- write vertex values -----*/
    DOFVector<int>::Iterator intPointIt(dc->getInterpPointInd(), USED_DOFS);
    DOFVector<double>::Iterator valueIt(dc->getValues(), USED_DOFS);
    DOFVector< ::std::list<WorldVector<double> > >::Iterator 
      coordIt(dc->getDofCoords(), USED_DOFS);

    int i;

    file << "vertex values: " << dc->getValues()->getName() << ::std::endl;

    for(intPointIt.reset(), valueIt.reset(), coordIt.reset();
	!intPointIt.end(); 
	++intPointIt, ++valueIt, ++coordIt) 
      {
	if(*intPointIt == -2) {
	  for(i=0; i < (int) coordIt->size(); i++) {
	    file << ::std::scientific << *valueIt << ::std::endl;
	  }
	}
      }

    file << ::std::endl << ::std::endl;

    /* ----- write interpolation values ----- */
    file << "interpolation values: " << dc->getValues()->getName() << ::std::endl;

    for(intPointIt.reset(), valueIt.reset();
	!intPointIt.end(); 
	++intPointIt, ++valueIt) 
      {
	if(*intPointIt >= 0) {
	  file << ::std::scientific << *valueIt << ::std::endl;
	}
      }

    file << ::std::endl << ::std::endl;

    /* ----- write interpolation points for each simplex */
    file << "element interpolation points: " << dc->getValues()->getName() << ::std::endl;

    ::std::list< ::std::list<int> >* interpPoints = dc->getInterpPoints();
    ::std::list< ::std::list<int> >::iterator it1;
    ::std::list<int>::iterator it2;

    for(it1 = interpPoints->begin(); it1 != interpPoints->end(); ++it1) {
      for(it2 = it1->begin(); it2 != it1->end(); ++it2) {
	file << (*it2) << " ";
      }
      file << ::std::endl;
    }
  
    file << ::std::endl;

    file.close();
  }


}
